#include <torch/extension.h>
#include <ATen/ATen.h>
#include <vector>

#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) CHECK_CUDA(x); CHECK_CONTIGUOUS(x)

#define TILE_DIM 32

__global__ void tiledMatMulReducedSync(const float* __restrict__ A, 
                                     const float* __restrict__ B, 
                                     float* __restrict__ C,
                                     int M, int D, int N) {
    int row = blockIdx.y * TILE_DIM + threadIdx.y;
    int col = blockIdx.x * TILE_DIM + threadIdx.x;
    
    __shared__ float tile_A[TILE_DIM][TILE_DIM];
    __shared__ float tile_B[TILE_DIM][TILE_DIM];
    
    float sum = 0.0f;
    
    // Register to store temporary results
    float reg_a, reg_b;
    
    for (int t = 0; t < (D + TILE_DIM - 1) / TILE_DIM; t++) {
        // Load data into registers first
        if (row < M && t * TILE_DIM + threadIdx.x < D) {
            reg_a = A[row * D + t * TILE_DIM + threadIdx.x];
        } else {
            reg_a = 0.0f;
        }
        
        if ((t * TILE_DIM + threadIdx.y) < D && col < N) {
            reg_b = B[(t * TILE_DIM + threadIdx.y) * N + col];
        } else {
            reg_b = 0.0f;
        }
        
        // Store from registers to shared memory
        tile_A[threadIdx.y][threadIdx.x] = reg_a;
        tile_B[threadIdx.y][threadIdx.x] = reg_b;
        
        // Single sync point after shared memory writes
        __syncthreads();
        
        // Compute partial sum using registers
        #pragma unroll
        for (int i = 0; i < TILE_DIM; i++) {
            sum += tile_A[threadIdx.y][i] * tile_B[i][threadIdx.x];
        }
        
        // Single sync point before next iteration
        __syncthreads();
    }
    
    // Write final result
    if (row < M && col < N) {
        C[row * N + col] = sum;
    }
}

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor clusters,
    torch::Tensor clusters2,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_running_mean,
    torch::Tensor bn_running_var,
    int64_t feature_size,
    int64_t cluster_size,
    bool is_training
) {
    CHECK_INPUT(x);
    CHECK_INPUT(clusters);
    CHECK_INPUT(clusters2);
    CHECK_INPUT(bn_weight);
    CHECK_INPUT(bn_bias);
    CHECK_INPUT(bn_running_mean);
    CHECK_INPUT(bn_running_var);

    int64_t B = x.size(0);
    int64_t N_points = x.size(1);
    int64_t D = feature_size;
    int64_t K = cluster_size;
    int64_t KplusG = clusters.size(1);

    auto x_reshaped = x.reshape({B * N_points, D});
    auto assignment = torch::empty({B * N_points, KplusG}, x.options());

    dim3 block(TILE_DIM, TILE_DIM);
    dim3 grid((KplusG + TILE_DIM - 1) / TILE_DIM, 
              (B * N_points + TILE_DIM - 1) / TILE_DIM);

    tiledMatMulReducedSync<<<grid, block>>>(
        x_reshaped.data_ptr<float>(),
        clusters.data_ptr<float>(),
        assignment.data_ptr<float>(),
        B * N_points, D, KplusG
    );

    assignment = torch::batch_norm(
        assignment,
        bn_weight,
        bn_bias,
        bn_running_mean,
        bn_running_var,
        is_training,
        0.1,
        1e-5,
        true
    );

    assignment = torch::softmax(assignment, 1);
    assignment = assignment.narrow(1, 0, K);
    assignment = assignment.reshape({B, N_points, K});

    auto a_sum = assignment.sum(1, /*keepdim=*/true);
    auto clusters2_exp = clusters2.expand({B, D, K});
    auto a = clusters2_exp * a_sum;

    assignment = assignment.transpose(1, 2);
    auto x_orig = x.reshape({B, N_points, D});
    auto vlad = torch::bmm(assignment, x_orig);

    vlad = vlad.transpose(1, 2);
    vlad = vlad - a;

    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));
    vlad = vlad.reshape({B, D * K});
    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));

    return vlad;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "NetVLAD forward with reduced synchronization");
}